/*!
 * \file      security_cache.hh
 * \brief     Header file for a certificates caching storage mechanism.
 *            It is used to store certificates received from neighbors and not present in the data base
 * \author    ETSI STF525
 * \copyright ETSI Copyright Notification
 *            No part may be reproduced except as authorized by written permission.
 *            The copyright and the foregoing restriction extend to reproduction in all media.
 *            All rights reserved.
 * \version   0.1
 */
#pragma once

#include <map>
#include <memory>
#include <string>
#include <vector>

#include "security_db_record.hh"

class OCTETSTRING; //! TITAN forward declaration
class INTEGER;     //! TITAN forward declaration
class CHARSTRING;  //! TITAN forward declaration

struct security_cache_comp {
  bool operator()(const OCTETSTRING &p_lhs, const OCTETSTRING &p_rhs) const;
};

/*!
 * \class security_cache
 * \brief This class provides security record description for in memory security database. In memory mens there is no disk load/save mechanism
 */
class security_cache {
  static bool fill_vector(OCTETSTRING &p_vector, const OCTETSTRING &p_org);

protected: /*! \protectedsection */
  // TODO Enforce with const security_cache_record, and const std::string
  std::map<std::string, const OCTETSTRING> _certificates_idx; //! List of the certificates indexed by the certificate identifier
  std::map<OCTETSTRING, const std::string, security_cache_comp>    _hashed_id8s;  //! List of the certificates indexed by the HashedId8
  std::map<std::string, std::unique_ptr<const security_db_record>> _certificates; //! List of the certificates indexed by the certificate identifier

  // FIXME In _certificates: relace key by hashed_id8s: std::map<OCTETSTRING&, std::unique_ptr<security_db_record>> _certificates;


public: /*! \publicsection */
  /*!
   * \brief Default ctor
   */
  security_cache();
  /*!
   * \brief Default private dtor
   */
  virtual ~security_cache();

  /*!
   * \fn int get_certificate_id(const OCTETSTRING &p_hashed_id8, std::string &p_certificate_id) const;
   * \brief Retrive the certificate name based of the specified HashedId8
   * \param[in] p_hashed_id8 The HashedId8 of the certificate
   * \param[out] p_certificate_name The certificate name
   * \return 0 on success, -1 otherwise
   */
  int get_certificate_id(const OCTETSTRING &p_hashed_id8, std::string &p_certificate_id) const;
  /*!
   * \fn int get_certificate(const OCTETSTRING &p_hashed_id8, OCTETSTRING &p_certificate);
   * \brief Retrive the specified certificate either from memory
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_certificate The raw certificate value
   * \return 0 on success, -1 otherwise
   */
  int get_certificate(const std::string &p_certificate_id, OCTETSTRING &p_certificate);
  /*!
   * \fn int get_certificate(const OCTETSTRING &p_hashed_id8, Ieee1609Dot2::CertificateBase &p_certificate);
   * \brief Retrive the specified certificate either from memory
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_certificate The Ieee1609Dot2 decoded certificate value
   * \return 0 on success, -1 otherwise
   */
  int get_certificate(const std::string &p_certificate_id, Ieee1609Dot2::CertificateBase &p_certificate);
  /*!
   * \fn int get_certificate(const OCTETSTRING &p_hashed_id3, Ieee1609Dot2::CertificateBase &p_certificate);
   * \brief Retrive the specified certificate either from memory
   * \param[in] p_certificate_id The HashedId3 of the certificate
   * \param[out] p_certificate The Ieee1609Dot2 decoded certificate value
   * \return 0 on success, -1 otherwise
   */
  int get_certificate(const OCTETSTRING &p_hashed_id3, Ieee1609Dot2::CertificateBase &p_certificate);
  /*!
   * \fn int get_certificate_hashed_id3(const OCTETSTRING &p_hashed_id3, std::string &p_certificate_id) const;
   * \brief Retrive the specified certificate either from memory
   * \param[in] p_hashed_id3 The HashedId3 of the certificate
   * \param[out] p_certificate_id The certificate name
   * \return 0 on success, -1 otherwise
   */
  int get_certificate_hashed_id3(const OCTETSTRING &p_hashed_id3, std::string &p_certificate_id) const;
  /*!
   * \fn int get_issuer(const std::string &p_certificate_id, OCTETSTRING &p_hashed_id_issuer);
   * \brief Retrive the issuer of the specified certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_hashed_id_issuer The HashedId8 of the isseur certificate
   * \return 0 on success, -1 otherwise
   */
  int get_issuer(const std::string &p_certificate_id, OCTETSTRING &p_hashed_id_issuer);
  /*!
   * \fn int get_hashed_id(const std::string &p_certificate_id, OCTETSTRING &p_hashed_id) const;
   * \brief Retrive the HashedId8 of the whole certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_hashed_id The HashedId8 of the whole certificate
   * \return 0 on success, -1 otherwise
   */
  int get_hashed_id(const std::string &p_certificate_id, OCTETSTRING &p_hashed_id) const;
  /*!
   * \fn int get_hash(const std::string &p_certificate_id, OCTETSTRING &p_hash);
   * \brief Retrive the hash (256 or 384) of the whole certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_hash The hash of the whole certificate
   * \return 0 on success, -1 otherwise
   */
  int get_hash(const std::string &p_certificate_id, OCTETSTRING &p_hash);
  /*!
   * \fn int get_hash_sha_256(const std::string &p_certificate_id, OCTETSTRING &p_hash);
   * \brief Retrive the SHA-256 of the whole certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_hash The SHA-256 of the whole certificate
   * \return 0 on success, -1 otherwise
   */
  int get_hash_sha_256(const std::string &p_certificate_id, OCTETSTRING &p_hash);
  /*!
   * \fn int get_private_key(const std::string &p_certificate_id, OCTETSTRING &p_private_key);
   * \brief Retrive the private key of the specified certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_private_key The private key
   * \return 0 on success, -1 otherwise
   */
  int get_private_key(const std::string &p_certificate_id, OCTETSTRING &p_private_key);
  /*!
   * \fn int get_public_keys(const std::string &p_certificate_id, OCTETSTRING &p_public_key_x, OCTETSTRING &p_public_key_y);
   * \brief Retrive the public keys (in uncompressed format) of the specified certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_public_key_x The X-coordinate of the public key
   * \param[out] p_public_key_y The Y-coordinate of the public key
   * \return 0 on success, -1 otherwise
   */
  int get_public_keys(const std::string &p_certificate_id, OCTETSTRING &p_public_key_x, OCTETSTRING &p_public_key_y);
  /*!
   * \fn int get_public_keys(const std::string &p_certificate_id, OCTETSTRING &p_public_key_x, OCTETSTRING &p_public_key_y);
   * \brief Retrive the public keys (in compressed format) of the specified certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_public_comp_key The X-coordinate of the public key
   * \param[out] p_comp_mode The compressed mode (2 for compress-y0, 3 otherwise)
   * \return 0 on success, -1 otherwise
   */
  int get_public_comp_key(const std::string &p_certificate_id, OCTETSTRING &p_public_comp_key, INTEGER &p_comp_mode);
  /*!
   * \fn int get_private_enc_key(const std::string &p_certificate_id, OCTETSTRING &p_private_enc_key);
   * \brief Retrive the private key for encryption of the specified certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_private_enc_key The private key for encryption
   * \return 0 on success, -1 otherwise
   */
  int get_private_enc_key(const std::string &p_certificate_id, OCTETSTRING &p_private_enc_key);
  /*!
   * \fn int get_public_enc_keys(const std::string &p_certificate_id, OCTETSTRING &p_public_enc_key_x, OCTETSTRING &p_public_enc_key_y);
   * \brief Retrive the public keys (in uncompressed format) for encryption of the specified certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_public_enc_key_x The X-coordinate of the public key for encryption
   * \param[out] p_public_enc_key_y The Y-coordinate of the public key for encryption
   * \return 0 on success, -1 otherwise
   */
  int get_public_enc_keys(const std::string &p_certificate_id, OCTETSTRING &p_public_enc_key_x, OCTETSTRING &p_public_enc_key_y);
  /*!
   * \fn int get_public_enc_comp_key(const std::string &p_certificate_id, OCTETSTRING &p_public_enc_comp_key, INTEGER &p_enc_comp_mode);
   * \brief Retrive the public keys (in compressed format) for encryption of the specified certificate
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_public_enc_comp_key The X-coordinate of the public key for encryption
   * \param[out] p_enc_comp_mode The compressed mode (2 for compress-y0, 3 otherwise)
   * \return 0 on success, -1 otherwise
   */
  int get_public_enc_comp_key(const std::string &p_certificate_id, OCTETSTRING &p_public_enc_comp_key, INTEGER &p_enc_comp_mode);

  virtual int store_certificate(const CHARSTRING &p_cert_id, const OCTETSTRING &p_cert, const OCTETSTRING &p_private_key, const OCTETSTRING &p_public_key_x,
                                const OCTETSTRING &p_public_key_y, const OCTETSTRING &p_public_compressed_key, const INTEGER &p_public_compressed_key_mode,
                                const OCTETSTRING &p_hash, const OCTETSTRING &p_hash_sha_256, const OCTETSTRING &p_hashid8, const OCTETSTRING &p_issuer,
                                const OCTETSTRING &p_private_enc_key, const OCTETSTRING &p_public_enc_key_x, const OCTETSTRING &p_public_enc_key_y,
                                const OCTETSTRING &p_public_enc_compressed_key, const INTEGER &p_public_enc_compressed_key_mode);

  /*!
   * \fn int int clear();
   * \brief Cleanup the cache
   * \return 0 on success, -1 otherwise
   * \virtual
   */
  virtual int clear();

  void dump() const;

private: /*! \privatesection */
  /*!
   * \fn int load_certificate(const std::string &p_certificate_id, const security_db_record** p_record);
   * \brief Retrive the specified certificate either from memory of from storage location
   * \param[in] p_certificate_id The certificate name
   * \param[out] p_record A reference to the certificate record
   * \return 0 on success, -1 otherwise
   */
  int load_certificate(const std::string &p_certificate_id, const security_db_record** p_record);

}; // End of class security_cache
